<?php

declare(strict_types=1);

namespace App\Controller\Admin\Ajax;

use App\Entity\Device;
use App\Entity\MonitoringGroup;
use App\Entity\MonitoringGroupMapping;
use App\Form\MonitoringGroupType;
use App\Repository\MonitoringGroupRepository;
use Doctrine\DBAL\Exception\UniqueConstraintViolationException;
use Doctrine\Persistence\ManagerRegistry;
use Symfony\Bundle\FrameworkBundle\Controller\AbstractController as Controller;
use Symfony\Component\Finder\Exception\AccessDeniedException;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;

/**
 * @Route("/admin/ajax/monitoring-group", name="ajax_monitoring_group_")
 */
class MonitoringGroupController extends Controller
{

    /**
     * Create new monitoring group
     * @Route("/create", methods={"POST"}, name="create")
     * @param Request $request
     * @param ManagerRegistry $managerRegistry
     * @return Response
     */
    public function createMonitoringGroupAction(Request $request, ManagerRegistry $managerRegistry): Response {

        $monitoringGroup = new MonitoringGroup();
        $form = $this->createForm(MonitoringGroupType::class, $monitoringGroup);
        $form->handleRequest($request);
        $errors = array();

        if ($form->isSubmitted()) {
            //$validator = $this->get('validator');
            //$errorsValidator = $validator->validate($student);
            $errorsValidator = [];
            foreach ($errorsValidator as $error) {
                array_push($errors, $error->getMessage());
            }

            if (count($errors) == 0) {

                $em =$managerRegistry->getManager();

                try{

                    $em->persist($monitoringGroup);
                    $em->flush();

                }catch (UniqueConstraintViolationException $e){

                    return new JsonResponse(array(
                        'code' => 400,
                        'message' => 'error',
                        'errors' => array('error' => "Monitoring Group already exists.")),
                        400);

                }catch (\Exception $e){

                    return new JsonResponse(array(
                        'code' => 400,
                        'message' => 'error',
                        'errors' => array('error' => "Unable to create the Monitoring Group.")),
                        400);

                }

                return new JsonResponse(array(
                    'code' => 200,
                    'message' => 'success',
                    'errors' => array('errors' => array(''))),
                    200);
            }

        }

        return new JsonResponse(array(
            'code' => 400,
            'message' => 'error',
            'errors' => array('errors' => $errors)),
            400);
    }

    /**
     * Get All Monitoring Groups
     * @Route("/all", methods={"GET"}, name="get_all")
     * @param Request $request
     * @param MonitoringGroupRepository $monitoringGroupRepository
     * @return Response
     */
    public function getAllMonitoringGroupsAction(Request $request, MonitoringGroupRepository $monitoringGroupRepository): Response {

        if ($request->isXmlHttpRequest()) {

            $monitoringGroups = $monitoringGroupRepository->getAllMonitoringGroups();
            $result = array();

            foreach ($monitoringGroups as $monitoringGroup) {
                $result[$monitoringGroup->getId()] = ['value' => $monitoringGroup->getName()];
            }

            return new JsonResponse($result);

        }

        return new Response('Bad request.', 400);

    }

    /**
     * Add Devices to the Monitoring Group
     * @Route("/add-devices", methods={"POST"}, name="devices_add")
     * @param Request $request
     * @param ManagerRegistry $managerRegistry
     * @return JsonResponse|Response
     */
    public function addDevicesToMonitoringGroup(Request $request, ManagerRegistry $managerRegistry): Response {

        if (!$request->isXmlHttpRequest()) {
            return new Response('Bad Request', 400);
        }

        if(isset($request->request)) {

            // Request has request data ?
            /*if (!isset($request->request))
                return new JsonResponse(array('status' => 'Error'),400);*/

            $items = (!is_null($request->get('items'))) ? (array) $request->get('items') : null;
            $mntGroupId = (!is_null($request->get('mntGroupId'))) ? (int) $request->get('mntGroupId') : null;

            if (is_null($items) || is_null($mntGroupId))
                return new JsonResponse(array('status' => 'Missing some data'),400);

            $em = $managerRegistry->getManager();

            $devices = $managerRegistry->getRepository(Device::class)->findById($items);
            $monitoringGroup = $managerRegistry->getRepository(MonitoringGroup::class)->findOneBy(['id' => $mntGroupId]);

            if($devices && $monitoringGroup){

                foreach ($devices as $device) {

                    $monitoringGroupMapping = new MonitoringGroupMapping();
                    $monitoringGroupMapping->setDevice($device);
                    $monitoringGroupMapping->setMonitoringGroup($monitoringGroup);
                    $em->persist($monitoringGroupMapping);

                }

                $em->flush();

                return new JsonResponse(array(
                    'code' => 200,
                    'message' => 'Devices sucessfully inserted to the group.',
                    'errors' => array('errors' => array(''))),
                    200);

            }

        }

        // If we reach this point, it means that something went wrong
        return new JsonResponse(array(
            'status' => 'Error',
            'message' => 'Error'),
            400);

    }

    /**
     * Remove Device form Monitoring Group
     * @Route("remove-device/{monitoringGroupId}/{deviceId}", methods={"GET"}, name="device_remove")
     * @param Request $request
     * @param int $monitoringGroupId
     * @param int $deviceId
     * @param ManagerRegistry $managerRegistry
     * @return JsonResponse|Response
     */
    public function deviceMonitoringGroupDelete(Request $request, int $monitoringGroupId, int $deviceId, ManagerRegistry $managerRegistry): Response {

        if (!$request->isXmlHttpRequest()) {
            return new Response('Bad Request', 400);
        }

        if(isset($request->request)) {

            if (!isset($monitoringGroupId) && !isset($deviceId))
                return new JsonResponse(array('status' => 'Error'),400);

            $monitoringGroup = $managerRegistry->getRepository(MonitoringGroup::class)->findOneBy(['id' => $monitoringGroupId]);
            $device = $managerRegistry->getRepository(Device::class)->findOneBy(['id' => $deviceId]);
            $monitoringGroupMapping = $managerRegistry->getRepository(MonitoringGroupMapping::class)->findOneBy(
                ['device' => $device, 'monitoringGroup' => $monitoringGroup]);

            if ($monitoringGroupMapping) {

                $em = $managerRegistry->getManager();
                $em->remove($monitoringGroupMapping);
                $em->flush();

            }

            return new JsonResponse([
                'code' => 200,
                'message' => 'Device from the Monitoring Group was removed.',
                'errors' => [],
                200]);

        }

        // If we reach this point, it means that something went wrong
        return new JsonResponse(array(
            'status' => 'Error',
            'message' => 'Error'),
            400);

    }

    /**
     * Remove devices from Monitoring Group
     * @Route("/remove-devices/{monitoringGroupId}", methods={"POST"}, name="devices_remove")
     * @param Request $request
     * @param int $monitoringGroupId
     * @param ManagerRegistry $managerRegistry
     * @return Response
     */
    public function removeDevicesFromMonitoringGroup(Request $request, int $monitoringGroupId, ManagerRegistry $managerRegistry): Response {

        if ($request->isXmlHttpRequest()) {

            $selectedDevices = $request->get('selectedDevices');
            //$token = $request->request->get('token');

            /*if (!$this->isCsrfTokenValid('multiselect', $token)) {
                throw new AccessDeniedException('The CSRF token is invalid.');
            }*/

            $em = $managerRegistry->getManager();
            $monitoringGroup = $managerRegistry->getRepository(MonitoringGroup::class)->findOneBy(['id' => $monitoringGroupId]);

            foreach ($selectedDevices as $device) {

                $device = $managerRegistry->getRepository(Device::class)->findOneBy(['id' => $device['id']]);
                $monitoringGroupMapping = $managerRegistry->getRepository(MonitoringGroupMapping::class)->findOneBy(
                    ['device' => $device, 'monitoringGroup' => $monitoringGroup]);

                if ($monitoringGroupMapping) {

                    $em = $managerRegistry->getManager();
                    $em->remove($monitoringGroupMapping);

                }
            }

            $em->flush();

            return new JsonResponse([
                'code' => 200,
                'message' => 'Devices from the monitoringGroup was removed.',
                'errors' => [],
                200]);

        }

        return new Response('Bad Request', 400);

    }

    /**
     * Delete Monitoring Group
     * @Route("/delete/{monitoringGroupId}", methods={"GET"}, name="delete")
     * @param Request $request
     * @param int $monitoringGroupId
     * @param ManagerRegistry $managerRegistry
     * @return Response
     */
    public function monitoringGroupDelete(Request $request, int $monitoringGroupId, ManagerRegistry $managerRegistry): Response {

        if ($request->isXmlHttpRequest()) {

            $monitoringGroup = $managerRegistry->getRepository(MonitoringGroup::class)->findOneBy(['id' => $monitoringGroupId]);

            if($monitoringGroup){

                $em = $managerRegistry->getManager();
                $em->remove($monitoringGroup);
                $em->flush();

            }

            return new JsonResponse(array(
                'code' => 200,
                'message' => 'Monitoring group deleted',
                'errors' => array('errors' => array(''))),
                200);

        }

        return new Response('Bad Request', 400);

    }

}